"""GMem (variable) object."""

from dataclasses import dataclass
from typing import Union

from aiovantage.config_client.xml_dataclass import xml_attribute, xml_element, xml_text

from .system_object import SystemObject


@dataclass
class GMem(SystemObject):
    """GMem (variable) object."""

    @dataclass
    class Tag:
        """GMem tag."""

        type: str = xml_text()
        object: bool = xml_attribute("object", default=False)

    @dataclass
    class Data:
        """GMem data."""

        fixed: bool = xml_attribute("Fixed", default=False)

    data: Data = xml_element("data")
    persistent: bool = xml_element("Persistent")
    tag: Tag = xml_element("Tag")

    def __post_init__(self) -> None:
        """Declare state attributes in post init."""
        self.value: Union[int, str, bool, None] = None

    @property
    def is_bool(self) -> bool:
        """Return True if GMem is boolean type."""
        return self.tag.type == "bool"

    @property
    def is_str(self) -> bool:
        """Return True if GMem is string type."""
        return self.tag.type == "Text"

    @property
    def is_int(self) -> bool:
        """Return True if GMem is integer type."""
        return self.tag.type in (
            "Delay",
            "DeviceUnits",
            "Level",
            "Load",
            "Number",
            "Seconds",
            "Task",
            "DegC",
        )

    @property
    def is_object_id(self) -> bool:
        """Return True if GMem is a pointer to an object."""
        return self.tag.object

    @property
    def is_fixed(self) -> bool:
        """Return True if GMem is a fixed point number."""
        return self.data.fixed



"""DDGColorLoad (RGB load) object."""

from dataclasses import dataclass

from .rgb_load import RGBLoad


@dataclass
class DDGColorLoad(RGBLoad):
    """DDGColorLoad (RGB load) object."""

    class Meta:
        """Meta class."""

        name = "Vantage.DDGColorLoad"



"""BlindGroup object."""

from dataclasses import dataclass
from typing import List, Optional

from aiovantage.config_client.xml_dataclass import xml_element

from .location_object import LocationObject


@dataclass
class BlindGroup(LocationObject):
    """BlindGroup object."""

    blind_ids: Optional[List[int]] = xml_element(
        "Blind", wrapper="BlindTable", default=None
    )



"""ILogin.Login method definition."""

from dataclasses import dataclass, field
from typing import ClassVar, Optional


@dataclass
class Login:
    """ILogin.Login method definition."""

    interface: ClassVar[str] = "ILogin"
    call: Optional["Login.Params"] = field(default=None)
    return_value: Optional[bool] = field(default=None, metadata={"name": "return"})

    @dataclass
    class Params:
        """ILogin.Login method parameters."""

        user: str = field(metadata={"name": "User"})
        password: str = field(metadata={"name": "Password"})



"""Somfy URTSI 2 blind group."""

from dataclasses import dataclass

from aiovantage.config_client.objects.blind_group import BlindGroup


@dataclass
class URTSI2Group(BlindGroup):
    """Somfy URTSI 2 blind group."""

    class Meta:
        """Meta class."""

        name = "Somfy.URTSI_2_Group_CHILD"



"""Wrapper for an asyncio connection to a Vantage controller."""

import asyncio
from ssl import CERT_NONE, SSLContext, create_default_context
from typing import ClassVar, Optional, Union

from .errors import ClientConnectionError, ClientTimeoutError


class BaseConnection:
    """Wrapper for an asyncio connection to a Vantage controller."""

    default_port: ClassVar[int]
    default_ssl_port: ClassVar[int]
    buffer_limit: ClassVar[int] = 2**16

    def __init__(
        self,
        host: str,
        port: Optional[int] = None,
        ssl: Union[SSLContext, bool] = True,
        conn_timeout: Optional[float] = None,
    ) -> None:
        """Initialize the connection."""
        self._host = host
        self._conn_timeout = conn_timeout
        self._reader: Optional[asyncio.StreamReader] = None
        self._writer: Optional[asyncio.StreamWriter] = None

        # Set up the SSL context
        self._ssl: Optional[SSLContext]
        if ssl is True:
            # We don't have a local issuer certificate to check against, and we'll be
            # connecting to an IP address so we can't check the hostname
            self._ssl = create_default_context()
            self._ssl.check_hostname = False
            self._ssl.verify_mode = CERT_NONE
        elif isinstance(ssl, SSLContext):
            self._ssl = ssl
        else:
            self._ssl = None

        # Set up the port
        self._port: int
        if port is None:
            self._port = self.default_ssl_port if ssl else self.default_port
        else:
            self._port = port

    async def open(self) -> None:
        """Open the connection."""
        # If we're already connected, do nothing
        if self._writer is not None and not self._writer.is_closing():
            return

        # Create the connection
        try:
            self._reader, self._writer = await asyncio.wait_for(
                asyncio.open_connection(
                    self._host,
                    self._port,
                    ssl=self._ssl,
                    limit=self.buffer_limit,
                ),
                timeout=self._conn_timeout,
            )
        except asyncio.TimeoutError as exc:
            raise ClientTimeoutError(
                f"Timeout connecting to {self._host}:{self._port}"
            ) from exc
        except OSError as exc:
            raise ClientConnectionError(
                f"Failed to connect to {self._host}:{self._port}"
            ) from exc

    def close(self) -> None:
        """Close the connection."""
        if self._writer is not None and not self._writer.is_closing():
            self._writer.close()
            self._writer = None

    @property
    def host(self) -> str:
        """Return the host."""
        return self._host

    @property
    def port(self) -> int:
        """Return the port."""
        return self._port

    @property
    def closed(self) -> bool:
        """Return whether the connection is closed."""
        return self._writer is None or self._writer.is_closing()

    async def write(self, message: str) -> None:
        """Send a plaintext message.

        Args:
            message: The message to send, as a string.
        """
        # Make sure we're connected
        if self._writer is None or self._writer.is_closing():
            raise ClientConnectionError("Client not connected.")

        # Send the request
        try:
            self._writer.write(message.encode())
            await self._writer.drain()
        except OSError as err:
            raise ClientConnectionError from err

    async def readuntil(self, separator: bytes, timeout: Optional[float] = None) -> str:
        """Read data until the separator is found or the optional timeout is reached.

        Args:
            separator: The separator to read until.
            timeout: The optional timeout in seconds.

        Returns:
            The data read, as a string.
        """
        # Make sure we're connected
        if self._reader is None or self.closed:
            raise ClientConnectionError("Client not connected.")

        # Read the response, with optional timeout
        try:
            data = await asyncio.wait_for(self._reader.readuntil(separator), timeout)
        except (OSError, asyncio.IncompleteReadError) as err:
            raise ClientConnectionError from err
        except asyncio.TimeoutError as err:
            raise ClientTimeoutError from err

        return data.decode()



"""DryContact object."""

from dataclasses import dataclass

from aiovantage.config_client.xml_dataclass import xml_element

from .location_object import LocationObject


@dataclass
class DryContact(LocationObject):
    """DryContact object."""

    parent_id: int = xml_element("Parent")

    def __post_init__(self) -> None:
        """Declare state attributes in post init."""
        self.triggered: bool = False



"""Fetch all Vantage controllers, and print any state changes."""

import argparse
import asyncio
import contextlib
import logging
from typing import Any, Dict

from aiovantage import Vantage, VantageEvent
from aiovantage.config_client.objects import Master

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


def callback(event: VantageEvent, obj: Master, data: Dict[str, Any]) -> None:
    """Print out any state changes."""
    if event == VantageEvent.OBJECT_ADDED:
        print(f"[Master added] '{obj.name}' ({obj.id})")

    elif event == VantageEvent.OBJECT_UPDATED:
        print(f"[Master updated] '{obj.name}' ({obj.id})")
        for attr in data.get("attrs_changed", []):
            print(f"    {attr} = {getattr(obj, attr)}")


async def main() -> None:
    """Run code example."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    async with Vantage(args.host, args.username, args.password) as vantage:
        # Subscribe to updates for all Vantage controllers
        vantage.masters.subscribe(callback)

        # Fetch all known Vantage controllers
        await vantage.masters.initialize()

        # Keep running for a while
        await asyncio.sleep(3600)


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""Helper functions for XML dataclasses."""

from dataclasses import field
from typing import Any, Type


def xml_attribute(name: str, **kwargs: Any) -> Any:
    """Create a dataclass field for an XML attribute."""
    metadata = {}
    metadata.update(kwargs.pop("metadata", {}))
    metadata.update({"name": name, "type": "Attribute"})

    return field(metadata=metadata, **kwargs)


def xml_element(name: str, **kwargs: Any) -> Any:
    """Create a dataclass field for an XML element."""
    metadata = {}
    metadata.update(kwargs.pop("metadata", {}))
    metadata.update({"name": name, "type": "Element"})

    if "wrapper" in kwargs:
        metadata.update({"wrapper": kwargs.pop("wrapper")})

    return field(metadata=metadata, **kwargs)


def xml_text(**kwargs: Any) -> Any:
    """Create a dataclass field for an XML text node."""
    metadata = {}
    metadata.update(kwargs.pop("metadata", {}))
    metadata.update({"type": "Text"})

    return field(metadata=metadata, **kwargs)


def xml_tag_from_class(cls: Type[Any]) -> str:
    """Get the XML tag name for a class."""
    meta = getattr(cls, "Meta", None)
    name = getattr(meta, "name", cls.__qualname__)

    return name



"""Light sensor object."""

from dataclasses import dataclass
from decimal import Decimal
from typing import Optional

from aiovantage.config_client.xml_dataclass import xml_element

from .sensor import Sensor


@dataclass
class LightSensor(Sensor):
    """Light sensor object."""

    parent_id: int = xml_element("Parent")

    def __post_init__(self) -> None:
        """Declare state attributes in post init."""
        self.level: Optional[Decimal] = None



"""IConfiguration.OpenFilter method definition."""

from dataclasses import dataclass, field
from typing import ClassVar, List, Optional


@dataclass
class OpenFilter:
    """IConfiguration.OpenFilter method definition."""

    interface: ClassVar[str] = "IConfiguration"
    call: Optional["OpenFilter.Params"] = field(default=None)
    return_value: Optional[int] = field(default=None, metadata={"name": "return"})

    @dataclass
    class Params:
        """IConfiguration.OpenFilter method parameters."""

        object_types: Optional[List[str]] = field(
            default=None, metadata={"name": "ObjectType", "wrapper": "Objects"}
        )
        xpath: Optional[str] = field(default=None, metadata={"name": "XPath"})



"""Interact with and control Vantage InFusion home automation controllers."""

__all__ = ["Vantage", "VantageEvent"]

import asyncio
from types import TracebackType
from typing import Callable, Optional, Type

from typing_extensions import Self

from aiovantage.command_client import CommandClient, EventStream
from aiovantage.config_client import ConfigClient
from aiovantage.config_client.objects import SystemObject
from aiovantage.controllers.anemo_sensors import AnemoSensorsController
from aiovantage.controllers.areas import AreasController
from aiovantage.controllers.base import EventCallback
from aiovantage.controllers.blind_groups import BlindGroupsController
from aiovantage.controllers.blinds import BlindsController
from aiovantage.controllers.buttons import ButtonsController
from aiovantage.controllers.dry_contacts import DryContactsController
from aiovantage.controllers.gmem import GMemController
from aiovantage.controllers.light_sensors import LightSensorsController
from aiovantage.controllers.load_groups import LoadGroupsController
from aiovantage.controllers.loads import LoadsController
from aiovantage.controllers.masters import MastersController
from aiovantage.controllers.modules import ModulesController
from aiovantage.controllers.omni_sensors import OmniSensorsController
from aiovantage.controllers.rgb_loads import RGBLoadsController
from aiovantage.controllers.stations import StationsController
from aiovantage.controllers.tasks import TasksController
from aiovantage.controllers.temperature_sensors import TemperatureSensorsController

from .events import VantageEvent


class Vantage:
    """Control a Vantage InFusion controller."""

    def __init__(
        self,
        host: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        *,
        use_ssl: bool = True,
        config_port: Optional[int] = None,
        command_port: Optional[int] = None,
    ) -> None:
        """Initialize the Vantage instance.

        Args:
            host: The hostname or IP address of the Vantage controller.
            username: The username to use for authentication.
            password: The password to use for authentication.
            use_ssl: Whether to use SSL for the connection.
            config_port: The port to use for the config client.
            command_port: The port to use for the command client.
        """
        # Set up clients
        self._config_client = ConfigClient(
            host, username, password, ssl=use_ssl, port=config_port
        )

        self._command_client = CommandClient(
            host, username, password, ssl=use_ssl, port=command_port
        )

        self._event_stream = EventStream(
            host, username, password, ssl=use_ssl, port=command_port
        )

        # Set up controllers
        self._anemo_sensors = AnemoSensorsController(self)
        self._areas = AreasController(self)
        self._blinds = BlindsController(self)
        self._blind_groups = BlindGroupsController(self)
        self._buttons = ButtonsController(self)
        self._dry_contacts = DryContactsController(self)
        self._gmem = GMemController(self)
        self._light_sensors = LightSensorsController(self)
        self._loads = LoadsController(self)
        self._load_groups = LoadGroupsController(self)
        self._masters = MastersController(self)
        self._modules = ModulesController(self)
        self._rgb_loads = RGBLoadsController(self)
        self._omni_sensors = OmniSensorsController(self)
        self._stations = StationsController(self)
        self._tasks = TasksController(self)
        self._temperature_sensors = TemperatureSensorsController(self)

    async def __aenter__(self) -> Self:
        """Return context manager."""
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exit context manager."""
        self.close()
        if exc_val:
            raise exc_val

    @property
    def config_client(self) -> ConfigClient:
        """Return the config client."""
        return self._config_client

    @property
    def command_client(self) -> CommandClient:
        """Return the command client."""
        return self._command_client

    @property
    def event_stream(self) -> EventStream:
        """Return the event stream."""
        return self._event_stream

    @property
    def anemo_sensors(self) -> AnemoSensorsController:
        """Return the AnemoSensors controller for managing anemo sensors."""
        return self._anemo_sensors

    @property
    def areas(self) -> AreasController:
        """Return the Areas controller for managing areas."""
        return self._areas

    @property
    def blinds(self) -> BlindsController:
        """Return the Blinds controller for managing blinds."""
        return self._blinds

    @property
    def blind_groups(self) -> BlindGroupsController:
        """Return the BlindGroups controller for managing groups of blinds."""
        return self._blind_groups

    @property
    def buttons(self) -> ButtonsController:
        """Return the Buttons controller for managing buttons."""
        return self._buttons

    @property
    def dry_contacts(self) -> DryContactsController:
        """Return the DryContacts controller for managing dry contacts."""
        return self._dry_contacts

    @property
    def light_sensors(self) -> LightSensorsController:
        """Return the LightSensors controller for managing light sensors."""
        return self._light_sensors

    @property
    def loads(self) -> LoadsController:
        """Return the Load controller for managing loads (lights, fans, etc)."""
        return self._loads

    @property
    def load_groups(self) -> LoadGroupsController:
        """Return the LoadGroup controller for managing groups of loads."""
        return self._load_groups

    @property
    def masters(self) -> MastersController:
        """Return the Masters controller for managing Vantage Controllers."""
        return self._masters

    @property
    def modules(self) -> ModulesController:
        """Return the Modules controller for managing modules."""
        return self._modules

    @property
    def gmem(self) -> GMemController:
        """Return the GMem controller for managing global memory."""
        return self._gmem

    @property
    def omni_sensors(self) -> OmniSensorsController:
        """Return the OmniSensors controller for managing generic sensors."""
        return self._omni_sensors

    @property
    def rgb_loads(self) -> RGBLoadsController:
        """Return the RGBLoads controller for managing RGB loads."""
        return self._rgb_loads

    @property
    def stations(self) -> StationsController:
        """Return the Stations controller for managing stations (keypads, etc)."""
        return self._stations

    @property
    def tasks(self) -> TasksController:
        """Return the Tasks controller for managing tasks."""
        return self._tasks

    @property
    def temperature_sensors(self) -> TemperatureSensorsController:
        """Return the TemperatureSensors controller for managing temperature sensors."""
        return self._temperature_sensors

    def close(self) -> None:
        """Close the clients."""
        self.config_client.close()
        self.command_client.close()
        self.event_stream.stop()

    async def initialize(self, fetch_state: bool = True) -> None:
        """Fetch all objects from the controllers.

        Args:
            fetch_state: Whether to also fetch the state of each object.
        """
        await asyncio.gather(
            self._anemo_sensors.initialize(fetch_state),
            self._areas.initialize(fetch_state),
            self._blinds.initialize(fetch_state),
            self._blind_groups.initialize(fetch_state),
            self._buttons.initialize(fetch_state),
            self._dry_contacts.initialize(fetch_state),
            self._gmem.initialize(fetch_state),
            self._light_sensors.initialize(fetch_state),
            self._loads.initialize(fetch_state),
            self._masters.initialize(fetch_state),
            self._modules.initialize(fetch_state),
            self._rgb_loads.initialize(fetch_state),
            self._omni_sensors.initialize(fetch_state),
            self._stations.initialize(fetch_state),
            self._tasks.initialize(fetch_state),
            self._temperature_sensors.initialize(fetch_state),
        )

    def subscribe(self, callback: EventCallback[SystemObject]) -> Callable[[], None]:
        """Subscribe to state changes for all objects.

        Args:
            callback: The callback to call when an object changes.

        Returns:
            A function to unsubscribe.
        """
        unsubscribes = [
            self.anemo_sensors.subscribe(callback),
            self.areas.subscribe(callback),
            self.blinds.subscribe(callback),
            self.blind_groups.subscribe(callback),
            self.buttons.subscribe(callback),
            self.dry_contacts.subscribe(callback),
            self.gmem.subscribe(callback),
            self.light_sensors.subscribe(callback),
            self.loads.subscribe(callback),
            self.masters.subscribe(callback),
            self.modules.subscribe(callback),
            self.rgb_loads.subscribe(callback),
            self.omni_sensors.subscribe(callback),
            self.stations.subscribe(callback),
            self.tasks.subscribe(callback),
            self.temperature_sensors.subscribe(callback),
        ]

        def unsubscribe() -> None:
            for unsub in unsubscribes:
                unsub()

        return unsubscribe



"""OmniSensor object."""

from dataclasses import dataclass
from decimal import Decimal
from typing import Union

from aiovantage.config_client.xml_dataclass import xml_attribute, xml_element

from .sensor import Sensor


@dataclass
class OmniSensor(Sensor):
    """OmniSensor object."""

    @dataclass
    class GetMethodType:
        """Omnisensor method information."""

        @dataclass
        class Formula:
            """OmniSensor type conversion information."""

            return_type: str = xml_attribute("ReturnType")
            level_type: str = xml_attribute("LevelType")

        formula: Formula = xml_element("Formula")
        method: str = xml_element("Method")
        method_hw: str = xml_element("MethodHW")

    get: GetMethodType = xml_element("Get")
    parent_id: int = xml_element("Parent")

    def __post_init__(self) -> None:
        """Declare state attributes in post init."""
        self.level: Union[int, Decimal, None] = None

    @property
    def is_current_sensor(self) -> bool:
        """Return True if the sensor is a current sensor."""
        return self.model == "Current"

    @property
    def is_power_sensor(self) -> bool:
        """Return True if the sensor is a power sensor."""
        return self.model == "Power"

    @property
    def is_temperature_sensor(self) -> bool:
        """Return True if the sensor is a temperature sensor."""
        return self.model == "Temperature"



"""Blind object."""

from dataclasses import dataclass
from decimal import Decimal
from typing import Optional

from aiovantage.config_client.xml_dataclass import xml_attribute, xml_element

from .location_object import LocationObject


@dataclass
class Blind(LocationObject):
    """Blind object."""

    parent_id: Optional[int] = xml_element("Parent", default=None)
    orientation: Optional[str] = xml_attribute("ShadeOrientation", default=None)
    type: Optional[str] = xml_attribute("ShadeType", default=None)

    def __post_init__(self) -> None:
        """Declare state attributes in post init."""
        self.position: Optional[Decimal] = None



"""Somfy RS-485 SDN 2.0 blind."""

from dataclasses import dataclass

from aiovantage.config_client.objects.blind import Blind


@dataclass
class RS485Shade(Blind):
    """Somfy RS-485 SDN 2.0 blind."""

    class Meta:
        """Meta class."""

        name = "Somfy.RS-485_Shade_CHILD"



"""LoadGroup object."""

from dataclasses import dataclass
from typing import List, Optional

from aiovantage.config_client.xml_dataclass import xml_element

from .location_object import LocationObject


@dataclass
class LoadGroup(LocationObject):
    """LoadGroup object."""

    load_ids: List[int] = xml_element("Load", wrapper="LoadTable")

    def __post_init__(self) -> None:
        """Declare state attributes in post init."""
        self.level: Optional[float] = None

    @property
    def is_on(self) -> bool:
        """Return True if the load is on."""
        return bool(self.level)



"""Power profile object."""

from dataclasses import dataclass

from aiovantage.config_client.xml_dataclass import xml_element

from .system_object import SystemObject


@dataclass
class PowerProfile(SystemObject):
    """Power Profile object."""

    min: float = xml_element("Min")
    max: float = xml_element("Max")
    adjust: int = xml_element("Adjust")
    freq: int = xml_element("Freq")
    inductive: bool = xml_element("Inductive")



"""Base class for all objects."""

from dataclasses import dataclass

from xsdata.models.datatype import XmlDateTime

from aiovantage.config_client.xml_dataclass import xml_attribute, xml_element


@dataclass
class SystemObject:
    """Base class for all objects."""

    id: int = xml_attribute("VID")
    master_id: int = xml_attribute("Master")
    mtime: XmlDateTime = xml_attribute("MTime")
    name: str = xml_element("Name")
    note: str = xml_element("Note")
    model: str = xml_element("Model")
    display_name: str = xml_element("DName")



"""ObjectChoice type definition."""

import inspect
from dataclasses import dataclass, field
from typing import Optional

import aiovantage.config_client.objects
from aiovantage.config_client.xml_dataclass import xml_tag_from_class

# Get all Vantage object from aiovantage.config_client.objects
ALL_OBJECT_TYPES = [
    item
    for _, item in inspect.getmembers(aiovantage.config_client.objects, inspect.isclass)
]


@dataclass
class ObjectChoice:
    """ObjectChoice type definition.

    Wildcard type that can be used to represent any object type.
    """

    id: Optional[int] = field(
        default=None, metadata={"name": "VID", "type": "Attribute"}
    )
    choice: Optional[object] = field(
        default=None,
        metadata={
            "type": "Wildcard",
            "choices": [
                {
                    "name": xml_tag_from_class(cls),
                    "type": cls,
                }
                for cls in ALL_OBJECT_TYPES
            ],
        },
    )



"""Button object."""

from dataclasses import dataclass

from aiovantage.config_client.xml_dataclass import xml_element

from .system_object import SystemObject


@dataclass
class Button(SystemObject):
    """Button object."""

    parent_id: int = xml_element("Parent")
    text1: str = xml_element("Text1")
    text2: str = xml_element("Text2")
    up_task_id: int = xml_element("Up")
    down_task_id: int = xml_element("Down")
    hold_task_id: int = xml_element("Hold")

    @property
    def has_task(self) -> bool:
        """Return True if button has a task assigned."""
        return any(
            task_id != 0
            for task_id in (
                self.up_task_id,
                self.down_task_id,
                self.hold_task_id,
            )
        )

    def __post_init__(self) -> None:
        """Declare state attributes in post init."""
        self.pressed: bool = False



"""Client for the Vantage Application Communication Interface (ACI) service.

The ACI service is an XML-based RPC service that Design Center uses to communicate
with Vantage InFusion Controllers. There are a number of "interfaces" exposed, each
with one or more "methods".

This service allows you to query the "configuration" of a Vantage system, for
example fetching a list of all the objects, getting a backup of the Design Center
XML, etc.

The service is exposed on port 2010 (SSL) by default, and on port 2001 (non-SSL) if
this port has been opened by the firewall on the controller.

The service is discoverable via mDNS as `_aci._tcp.local` and/or
`_secure_aci._tcp.local`.
"""

import asyncio
import logging
from ssl import SSLContext
from types import TracebackType
from typing import Optional, Type, Union
from xml.etree import ElementTree

from typing_extensions import Self
from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.parsers.config import ParserConfig
from xsdata.formats.dataclass.parsers.handlers import XmlEventHandler
from xsdata.formats.dataclass.serializers import XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig

from aiovantage.config_client.methods import Call, Method, Return
from aiovantage.config_client.methods.introspection import GetVersion
from aiovantage.config_client.methods.login import Login
from aiovantage.connection import BaseConnection
from aiovantage.errors import ClientResponseError, LoginFailedError, LoginRequiredError


class ConfigConnection(BaseConnection):
    """Connection to a Vantage ACI server."""

    default_port = 2001
    default_ssl_port = 2010
    buffer_limit = 2**20


class ConfigClient:
    """Client for the Vantage Application Communication Interface (ACI) service.

    This client handles connecting to the ACI service, authenticating, and the
    serialization/deserialization of XML requests and responses.
    """

    def __init__(
        self,
        host: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        *,
        ssl: Union[SSLContext, bool] = True,
        port: Optional[int] = None,
        conn_timeout: float = 30,
        read_timeout: float = 60,
    ) -> None:
        """Initialize the client."""
        self._connection = ConfigConnection(host, port, ssl, conn_timeout)
        self._username = username
        self._password = password
        self._read_timeout = read_timeout

        self._serializer = XmlSerializer(
            config=SerializerConfig(xml_declaration=False),
        )

        self._parser = XmlParser(
            config=ParserConfig(fail_on_unknown_properties=False),
            handler=XmlEventHandler,
        )
        self._connection_lock = asyncio.Lock()
        self._request_lock = asyncio.Lock()
        self._logger = logging.getLogger(__name__)

    async def __aenter__(self) -> Self:
        """Return context manager."""
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exit context manager."""
        self.close()
        if exc_val:
            raise exc_val

    def close(self) -> None:
        """Close the connection to the ACI service."""
        self._connection.close()

    async def request(
        self,
        method_cls: Type[Method[Call, Return]],
        params: Optional[Call] = None,
        connection: Optional[ConfigConnection] = None,
    ) -> Optional[Return]:
        """Marshall a request, send it to the ACI service, and return a parsed object.

        Args:
            method_cls: The request method class to use.
            params: The parameters to pass to the method.
            connection: The connection to use, if not the default.

        Returns:
            The parsed response object
        """
        # Open the connection if it's closed
        conn = connection or await self.get_connection()

        # Build the method object
        method = method_cls()
        method.call = params

        # Render the method object to XML with xsdata
        request = (
            f"<{method.interface}>"
            + self._serializer.render(method)
            + f"</{method.interface}>"
        )
        self._logger.debug("Sending request: %s", request)

        # Send the request and read the response
        async with self._request_lock:
            await conn.write(request)
            response = await conn.readuntil(
                f"</{method.interface}>\n".encode(), timeout=self._read_timeout
            )
        self._logger.debug("Received response: %s", response)

        # Parse the XML doc
        root = ElementTree.fromstring(response)

        # Response root must match the tag of the request
        if root.tag != method.interface:
            raise ClientResponseError(
                f"Response '{root.tag}' does not match request '{method.interface}'"
            )

        # Responses must contain the method element and a return element
        method_el = root.find(f"./{method_cls.__name__}")
        return_el = root.find(f"./{method_cls.__name__}/return")
        if method_el is None or return_el is None:
            raise ClientResponseError("Response is missing method or return element")

        # Return None if the method has empty return element.
        # This can happen when the client is not logged in, or when the method returns
        # no results, eg. IConfiguration.GetFilterResults.
        if return_el.text is None and len(return_el) == 0:
            return None

        # Parse the method element with xsdata
        method = self._parser.parse(method_el, method_cls)
        if method.return_value is None:
            return None

        return method.return_value

    async def get_connection(self) -> ConfigConnection:
        """Get a connection to the ACI service."""
        async with self._connection_lock:
            if self._connection.closed:
                await self._connection.open()

                # Ensure the connection is authenticated, if required
                if self._username and self._password:
                    # Log in if we have credentials
                    success = await self.request(
                        Login,
                        Login.Params(self._username, self._password),
                        self._connection,
                    )

                    if not success:
                        raise LoginFailedError("Login failed, bad username or password")
                else:
                    # Check if login is required if we don't have credentials
                    version = await self.request(
                        GetVersion,
                        connection=self._connection,
                    )

                    if version is None:
                        raise LoginRequiredError(
                            "Login required, but no credentials were provided"
                        )

            return self._connection



"""Base class for all station objects."""

from dataclasses import dataclass

from aiovantage.config_client.xml_dataclass import xml_element

from .location_object import LocationObject


@dataclass
class StationObject(LocationObject):
    """Base class for all station objects."""

    serial_number: str = xml_element("SerialNumber")
    bus_id: int = xml_element("Bus")



"""IConfiguration.GetFilterResults method definition."""

from dataclasses import dataclass, field
from typing import ClassVar, List, Optional

from aiovantage.config_client.methods.types import ObjectChoice


@dataclass
class GetFilterResults:
    """IConfiguration.GetFilterResults method definition."""

    interface: ClassVar[str] = "IConfiguration"
    call: Optional["GetFilterResults.Params"] = field(default=None)
    return_value: Optional[List[ObjectChoice]] = field(
        default=None, metadata={"name": "Object", "wrapper": "return"}
    )

    @dataclass
    class Params:
        """IConfiguration.GetFilterResults method parameters."""

        h_filter: int = field(metadata={"name": "hFilter"})
        count: int = field(default=50, metadata={"name": "Count"})
        whole_object: bool = field(default=True, metadata={"name": "WholeObject"})



"""Load object."""

from dataclasses import dataclass
from typing import Optional

from aiovantage.config_client.xml_dataclass import xml_element

from .location_object import LocationObject


@dataclass
class Load(LocationObject):
    """Load object."""

    load_type: str = xml_element("LoadType")
    parent_id: int = xml_element("Parent")
    power_profile_id: int = xml_element("PowerProfile")

    def __post_init__(self) -> None:
        """Declare state attributes in post init."""
        self.level: Optional[float] = None

    @property
    def is_relay(self) -> bool:
        """Return True if the load is a relay."""
        return self.load_type in (
            "High Voltage Relay",
            "Low Voltage Relay",
        )

    @property
    def is_motor(self) -> bool:
        """Return True if the load is a motor."""
        return self.load_type == "Motor"

    @property
    def is_light(self) -> bool:
        """Return True if the load is a light."""
        return not (self.is_relay or self.is_motor)

    @property
    def is_dimmable(self) -> bool:
        """Return True if the load is dimmable."""
        return not self.load_type.endswith("non-Dim")

    @property
    def is_on(self) -> bool:
        """Return True if the load is on."""
        return bool(self.level)



"""Somfy URTSI 2 blind."""

from dataclasses import dataclass

from aiovantage.config_client.objects.blind import Blind


@dataclass
class URTSI2Shade(Blind):
    """Somfy URTSI 2 blind."""

    class Meta:
        """Meta class."""

        name = "Somfy.URTSI_2_Shade_CHILD"



"""RGB load object."""

from dataclasses import dataclass
from enum import Enum, IntEnum
from typing import Optional, Tuple

from aiovantage.config_client.xml_dataclass import xml_element

from .location_object import LocationObject

# This isn't strictly a Vantage object type, but this helps us avoid
# code duplication in RGBLoad-like objects.


@dataclass
class RGBLoad(LocationObject):
    """RGB load object."""

    class RGBChannel(IntEnum):
        """RGB channel."""

        RED = 0
        GREEN = 1
        BLUE = 2
        WHITE = 3

    class HSLAttribute(IntEnum):
        """HSL attribute."""

        HUE = 0
        SATURATION = 1
        LEVEL = 2

    class ColorType(Enum):
        """Color type."""

        RGB = "RGB"
        RGBW = "RGBW"
        HSL = "HSL"
        HSIC = "HSIC"
        CCT = "CCT"
        COLOR_CHANNEL = "Color Channel"

    color_type: ColorType = xml_element("ColorType")
    min_temp: int = xml_element("MinTemp")
    max_temp: int = xml_element("MaxTemp")
    parent_id: int = xml_element("Parent")

    def __post_init__(self) -> None:
        """Declare state attributes in post init."""
        self.hsl: Optional[Tuple[int, int, int]] = None
        self.rgb: Optional[Tuple[int, int, int]] = None
        self.rgbw: Optional[Tuple[int, int, int, int]] = None
        self.level: Optional[int] = None
        self.color_temp: Optional[int] = None

    @property
    def is_on(self) -> bool:
        """Return True if the load is on."""
        return bool(self.level)

    @property
    def is_rgb(self) -> bool:
        """Return True if the load is an RGB(W) load."""
        return self.color_type in (
            RGBLoad.ColorType.RGB,
            RGBLoad.ColorType.RGBW,
            RGBLoad.ColorType.HSL,
        )

    @property
    def is_cct(self) -> bool:
        """Return True if the load is a CCT load."""
        return self.color_type == RGBLoad.ColorType.CCT



"""Base class for system objects in an area."""

from dataclasses import dataclass

from aiovantage.config_client.xml_dataclass import xml_element

from .system_object import SystemObject


@dataclass
class LocationObject(SystemObject):
    """Base class for system objects in an area."""

    area_id: int = xml_element("Area")
    location: str = xml_element("Location")



"""Task object."""

from dataclasses import dataclass
from typing import Optional

from .system_object import SystemObject


@dataclass
class Task(SystemObject):
    """Task object."""

    def __post_init__(self) -> None:
        """Declare state attributes in post init."""
        self.is_running: Optional[bool] = None
        self.state: Optional[bool] = None



"""Station bus object."""

from dataclasses import dataclass

from aiovantage.config_client.xml_dataclass import xml_element

from .system_object import SystemObject


@dataclass
class StationBus(SystemObject):
    """Station bus object."""

    parent_id: int = xml_element("Parent")



"""Master (controller) object."""

from dataclasses import dataclass
from typing import Optional

from aiovantage.config_client.xml_dataclass import xml_element

from .system_object import SystemObject


@dataclass
class Master(SystemObject):
    """Master (controller) object."""

    number: int = xml_element("Number")
    volts: float = xml_element("Volts")
    amps: float = xml_element("Amps")
    module_count: int = xml_element("ModuleCount")
    serial_number: int = xml_element("SerialNumber")

    def __post_init__(self) -> None:
        """Post init hook."""
        self.last_updated: Optional[int] = None
        self.firmware_version: Optional[str] = None



"""Somfy RS-485 SDN 2.0 blind group."""

from dataclasses import dataclass

from aiovantage.config_client.objects.blind_group import BlindGroup


@dataclass
class RS485Group(BlindGroup):
    """Somfy RS-485 SDN 2.0 blind group."""

    class Meta:
        """Meta class."""

        name = "Somfy.RS-485_Group_CHILD"
